# This file is part of sale_kit module for Tryton.
# The COPYRIGHT file at the top level of this repository contains
# the full copyright notices and license terms.
from decimal import Decimal
from trytond.model import fields
from trytond.pool import Pool, PoolMeta
from trytond.pyson import Equal, Eval
from trytond.transaction import Transaction

__all__ = ['SaleLine']


class SaleLine(metaclass=PoolMeta):
    __name__ = 'sale.line'
    kit_depth = fields.Integer('Depth', required=True,
        help='Depth of the line if it is part of a kit.')
    kit_parent_line = fields.Many2One('sale.line', 'Parent Kit Line',
        help='The kit that contains this product.')
    kit_child_lines = fields.One2Many('sale.line', 'kit_parent_line',
        'Lines in the kit', help='Subcomponents of the kit.')

    @classmethod
    def __setup__(cls):
        super(SaleLine, cls).__setup__()
        required = (~(Eval('kit_parent_line', False))
            and (Equal(Eval('type'), 'line')))
        cls.unit_price.states['required'] = required

    @classmethod
    def default_kit_depth(cls):
        return 0

    def _fill_line_from_kit_line(self, kit_line, line):
        ProductUom = Pool().get('product.uom')

        self.product = kit_line.product
        if kit_line.unit.category.id != line.unit.category.id:
            quantity = kit_line.quantity * line.quantity
        else:
            quantity = ProductUom.compute_qty(
                    kit_line.unit, kit_line.quantity, line.unit
                    ) * line.quantity
        self.quantity = quantity
        self.unit = kit_line.unit
        self.type = 'line'
        self.kit_parent_line = line

    @classmethod
    def explode_kit(cls, lines):
        '''
        Walks through the Kit tree in depth-first order and returns
        a sorted list with all the components of the product.
        If no product on Sale Line avoid to try explode kits
        '''
        def max_sequence(sale_lines):
            return max([sl.sequence for sl in sale_lines] +
                [max_sequence(sl.kit_child_lines) for sl in sale_lines
                    if sl.kit_child_lines])

        Product = Pool().get('product.product')

        sale_discount = hasattr(cls, 'gross_unit_price')

        sequence = lines[0].sequence if lines and lines[0].sequence else 1
        to_write, to_create = [], []
        for line in lines:
            if line.sequence != sequence and to_create:
                line.sequence = sequence
            sequence += 1
            depth = line.kit_depth + 1
            if (line.product and line.product.kit and line.product.kit_lines
                    and line.product.explode_kit_in_sales):
                kit_lines = line.product.kit_lines
                kit_lines = list(zip(kit_lines, [depth] * len(list(kit_lines))))
                while kit_lines:
                    kit_line = kit_lines.pop(0)
                    depth = kit_line[1]
                    kit_line = kit_line[0]
                    product = kit_line.product

                    sale_line = cls()
                    for key, value in list(sale_line.default_get(
                            list(sale_line._fields.keys()),
                            with_rec_name=False).items()):
                        if value is not None:
                            setattr(sale_line, key, value)
                    # add party/sid when create new line with
                    # sale_line_standalone or galatea_esale
                    if hasattr(line, 'party'):
                        sale_line.party = line.party
                    if hasattr(line, 'sid'):
                        sale_line.sid = line.sid
                    sale_line.sale = line.sale
                    sale_line._fill_line_from_kit_line(kit_line, line)
                    sale_line.sequence = sequence
                    sale_line.on_change_product()
                    sale_line.kit_depth = depth

                    if kit_line.get_sale_price():
                        with Transaction().set_context(
                                sale_line._get_context_sale_price()):
                            prices = Product.get_sale_price(
                                [product], line.quantity)
                            unit_price = prices[product.id]
                            digits = cls.unit_price.digits[1]
                            unit_price = unit_price.quantize(
                                Decimal(1) / 10 ** digits)
                    else:
                        unit_price = Decimal('0.0')

                    # Compatibility with sale_discount module
                    if sale_discount:
                        sale_line.gross_unit_price = unit_price
                        if line.discount:
                            sale_line.discount = line.discount
                        sale_line.on_change_discount()
                    else:
                        sale_line.unit_price = unit_price

                    to_create.append(sale_line._save_values)
                    if product.kit_lines:
                        product_kit_lines = product.kit_lines
                        product_kit_lines = zip(product_kit_lines,
                            [depth + 1] * len(list(product_kit_lines)))
                        kit_lines = product_kit_lines + kit_lines
                    sequence += 1
                if not line.product.kit_fixed_list_price and line.unit_price:
                    line.unit_price = Decimal('0.0')
            elif (line.product and line.product.kit_lines and
                    not line.product.kit_fixed_list_price):
                with Transaction().set_context(
                        line._get_context_sale_price()):
                    prices = Product.get_sale_price(
                        [line.product], line.quantity)
                    unit_price = prices[line.product.id]

                # Compatibility with sale_discount module
                if sale_discount:
                    line.gross_unit_price = unit_price
                    line.on_change_discount()
                else:
                    # Avoid modifing when not required
                    if line.unit_price != unit_price:
                        line.unit_price = unit_price
            if line._save_values:
                to_write.extend(([line], line._save_values))
        if to_write:
            cls.write(*to_write)
        # Call super create to avoid recursion error
        return super(SaleLine, cls).create(to_create)

    @classmethod
    def create(cls, values):
        lines = super(SaleLine, cls).create(values)
        if (Transaction().context.get('explode_kit', True)
                and not Transaction().context.get('standalone', False)):
            lines.extend(cls.explode_kit(lines))
        return lines

    def get_kit_lines(self, kit_line=None):
        res = []
        if kit_line:
            childs = kit_line.kit_child_lines
        else:
            childs = self.kit_child_lines
        for kit_line in childs:
            res.append(kit_line)
            res += self.get_kit_lines(kit_line)
        return res

    @classmethod
    def write(cls, *args):
        actions = iter(args)
        to_write, to_reset, to_delete = [], [], []
        if (Transaction().context.get('explode_kit', True)
                and not Transaction().context.get('standalone', False)):
            for lines, values in zip(actions, actions):
                reset_kit = False
                # TODO Explode kit when add new line from standalone
                if ('product' in values or 'quantity' in values
                        or 'unit' in values):
                    reset_kit = True
                lines = lines[:]
                if reset_kit:
                    for line in lines:
                        to_delete += line.get_kit_lines()
                    lines = list(set(lines) - set(to_delete))
                    to_reset.extend(lines)
                to_write.extend((lines, values))
        else:
            to_write = args
        super(SaleLine, cls).write(*to_write)
        if to_delete:
            cls.delete(to_delete)
        if to_reset:
            cls.explode_kit(to_reset)

    @classmethod
    def copy(cls, lines, default=None):
        if default is None:
            default = {}
        default['kit_child_lines'] = []
        lines = [x for x in lines if not x.kit_parent_line]
        new_lines = super(SaleLine, cls).copy(lines, default=default)
        return new_lines

    def get_invoice_line(self):
        lines = super(SaleLine, self).get_invoice_line()
        for line in lines:
            line.sequence = self.sequence
        return lines




# The COPYRIGHT file at the top level of this repository contains
# the full copyright notices and license terms.
import unittest
import doctest

from trytond.tests.test_tryton import ModuleTestCase
from trytond.tests.test_tryton import suite as test_suite
from trytond.tests.test_tryton import doctest_teardown, doctest_checker


class SaleKitTestCase(ModuleTestCase):
    'Test Sale Kit module'
    module = 'sale_kit'
    extras = ['purchase']


def suite():
    suite = test_suite()
    suite.addTests(unittest.TestLoader().loadTestsFromTestCase(
            SaleKitTestCase))
    suite.addTests(doctest.DocFileSuite('scenario_sale_kit.rst',
            tearDown=doctest_teardown, encoding='utf-8',
            optionflags=doctest.REPORT_ONLY_FIRST_FAILURE,
            checker=doctest_checker))
    return suite



# The COPYRIGHT file at the top level of this repository contains the full
# copyright notices and license terms.
from trytond.pool import Pool
from . import product
from . import sale

__all__ = ['register']


def register():
    Pool.register(
        product.Product,
        product.ProductKitLine,
        product.ProductSupplier,
        sale.Sale,
        sale.SaleLine,
        module='sale_kit', type_='model')



# The COPYRIGHT file at the top level of this repository contains
# the full copyright notices and license terms.
from decimal import Decimal
from itertools import groupby, chain
from functools import partial

from trytond.model import fields
from trytond.pool import Pool, PoolMeta
from trytond.pyson import Equal, Eval
from trytond.transaction import Transaction
from trytond.modules.sale.exceptions import (
    SaleValidationError, SaleQuotationError, PartyLocationError)


class Sale(metaclass=PoolMeta):
    __name__ = 'sale.sale'

    def get_moves_for_line(self, line, shipment_type):
        '''
        Get the moves for a sale line

        - needs patch sale_issue2762_refactor_create_shipment.diff
        '''
        moves = {}
        if line.product and line.product.kit:
            kit_moves = line.get_moves_for_kit(shipment_type)
            if kit_moves:
                i = 0
                for move in kit_moves:
                    # Provide a unique line id to avoid collision of
                    # multiple moves for the same line (must be translated
                    # back later in _group_shipment_key to enable correct
                    # grouping per sale line id)
                    move_id = '%s_%ssale_kit' % (line.id, i)
                    moves[move_id] = move
                    i += 1
        else:
            moves = super(Sale, self).get_moves_for_line(line, shipment_type)
        return moves

    def _group_shipment_key(self, moves, move):
        '''
        Check if the provided line_id was encoded by this module
        and translate it back.
        '''
        line_id, move_obj = move
        if isinstance(line_id, str) and line_id.endswith('sale_kit'):
            line_id = int(line_id.split('_')[0])
            move = (line_id, move_obj)
        return super(Sale, self)._group_shipment_key(moves, move)

    _group_return_key = _group_shipment_key


class SaleLine(metaclass=PoolMeta):
    __name__ = 'sale.line'
    kit_depth = fields.Integer('Depth', required=True,
        help='Depth of the line if it is part of a kit.')
    kit_parent_line = fields.Many2One('sale.line', 'Parent Sale Line',
        help='The sale line that contains this product.')
    kit_child_lines = fields.One2Many('sale.line', 'kit_parent_line',
        'Kit Components', help='The components of the kit')

    @classmethod
    def __setup__(cls):
        super(SaleLine, cls).__setup__()
        required = (~(Eval('kit_parent_line', False))
            and (Equal(Eval('type'), 'line')))
        cls.unit_price.states['required'] = required

    @staticmethod
    def default_kit_depth():
        return 0

    @fields.depends('product', 'description', 'sale')
    def on_change_product(self):
        super(SaleLine, self).on_change_product()
        prefix = ' >'
        if self.description and prefix not in self.description:
            if (self.product and self.product.kit and self.product.kit_lines
                    and self.product.components_in_description):
                party = None
                party_context = {}
                if self.sale and self.sale.party:
                    party = self.sale.party
                    if party.lang:
                        party_context['language'] = party.lang.code
                with Transaction().set_context(party_context):
                    desc = ''
                    for component in self.product.kit_lines:
                        desc += '\n' + ' '.join([prefix,
                                str(component.quantity),
                                component.unit.rec_name,
                                component.product.rec_name])
                self.description += '\n' + desc

    def _fill_line_from_kit_line(self, kit_line, line):
        ProductUom = Pool().get('product.uom')

        self.product = kit_line.product
        if kit_line.unit.category.id != line.unit.category.id:
            quantity = kit_line.quantity * line.quantity
        else:
            quantity = ProductUom.compute_qty(
                    kit_line.unit, kit_line.quantity, line.unit
                    ) * line.quantity
        self.quantity = quantity
        self.unit = kit_line.unit
        self.type = 'line'
        self.kit_parent_line = line
        self.description = ''

    @classmethod
    def explode_kit(cls, lines):
        '''
        Walks through the Kit tree in depth-first order and returns
        a sorted list with all the components of the product.
        If no product on Sale Line avoid to try explode kits
        '''
        #def max_sequence(sale_lines):
        #    return max([sl.sequence for sl in sale_lines] +
        #        [max_sequence(sl.kit_child_lines) for sl in sale_lines
        #            if sl.kit_child_lines])

        Product = Pool().get('product.product')

        sale_discount = hasattr(cls, 'gross_unit_price')

        sequence = lines[0].sequence if lines and lines[0].sequence else 1
        to_write, to_create = [], []
        for line in lines:
            if line.sequence != sequence and to_create:
                line.sequence = sequence
            sequence += 1
            depth = line.kit_depth + 1
            if (line.product and line.product.kit and line.product.kit_lines
                    and line.product.explode_kit_in_sales):
                kit_lines = list(line.product.kit_lines)
                kit_lines = list(zip(kit_lines, [depth] * len(kit_lines)))
                while kit_lines:
                    kit_line = kit_lines.pop(0)
                    depth = kit_line[1]
                    kit_line = kit_line[0]
                    product = kit_line.product

                    sale_line = cls()
                    for key, value in sale_line.default_get(
                            list(sale_line._fields.keys()),
                            with_rec_name=False).items():
                        if value is not None:
                            setattr(sale_line, key, value)
                    if hasattr(line, 'party'):
                        sale_line.party = line.party
                    if hasattr(line, 'sid'):
                        sale_line.sid = line.sid
                    sale_line.sale = line.sale
                    sale_line._fill_line_from_kit_line(kit_line, line)
                    sale_line.sequence = sequence
                    sale_line.on_change_product()
                    sale_line.kit_depth = depth
                    sale_line.description = ('%s%s' %
                        ('> ' * depth, sale_line.description)
                        if sale_line.description else ' ')

                    if kit_line.get_sale_price():
                        with Transaction().set_context(
                                sale_line._get_context_sale_price()):
                            prices = Product.get_sale_price(
                                [product], line.quantity)
                            unit_price = prices[product.id]
                            digits = cls.unit_price.digits[1]
                            unit_price = unit_price.quantize(
                                Decimal(1) / 10 ** digits)
                    else:
                        unit_price = Decimal('0.0')

                    # Compatibility with sale_discount module
                    if sale_discount:
                        sale_line.gross_unit_price = unit_price
                        if line.discount:
                            sale_line.discount = line.discount
                        sale_line.on_change_discount()
                    else:
                        sale_line.unit_price = unit_price

                    to_create.append(sale_line._save_values)
                    if product.kit_lines:
                        product_kit_lines = list(product.kit_lines)
                        product_kit_lines = list(zip(product_kit_lines,
                            [depth + 1] * len(product_kit_lines)))
                        kit_lines = product_kit_lines + kit_lines
                    sequence += 1
                if not line.product.kit_fixed_list_price and line.unit_price:
                    line.unit_price = Decimal('0.0')
            elif (line.product and line.product.kit_lines and
                    not line.product.kit_fixed_list_price):
                with Transaction().set_context(
                        line._get_context_sale_price()):
                    prices = Product.get_sale_price(
                        [line.product], line.quantity)
                    unit_price = prices[line.product.id]

                # Compatibility with sale_discount module
                if sale_discount:
                    line.gross_unit_price = unit_price
                    line.on_change_discount()
                else:
                    # Avoid modifing when not required
                    if line.unit_price != unit_price:
                        line.unit_price = unit_price
            if line._save_values:
                to_write.extend(([line], line._save_values))
        if to_write:
            cls.write(*to_write)
        # Call super create to avoid recursion error
        return super(SaleLine, cls).create(to_create)

    def _get_shipped_quantity(self, shipment_type):
        '''
        Override SaleLine._get_shipped_quantity for not exploded kits
        to get correct moves.
        '''
        if (self.type == 'line' and self.product and self.product.kit
                and self.product.kit_lines
                and not self.product.explode_kit_in_sales):
            kit_quantity = self.quantity
            for line in self.product.kit_lines:
                to_ship = kit_quantity * line.quantity
                shipped = self._get_shipped_quantity_for_kit_line(line)
                if shipped < to_ship:
                    kit_quantity = 0
                    break
            return kit_quantity
        else:
            return super(SaleLine, self)._get_shipped_quantity(shipment_type)

    def _get_shipped_quantity_for_kit_line(self, kit_line):
        pool = Pool()
        Uom = pool.get('product.uom')

        skips = set(m for m in self.moves_recreated)
        shipped = 0
        for move in self.moves:
            if move not in skips:
                if kit_line.product == move.product:
                    shipped += Uom.compute_qty(move.uom,
                        move.quantity, kit_line.unit)
        return shipped

    def get_moves_for_kit_org(self, shipment_type):
        '''
        Walk through the kit tree in depth-first order, get
        a sorted list with all the components of the kit and
        create moves for the kit components.

        s.a. sale/sale.py/get_move
        s.a. sale_kit/product.py/KitLine/get_move
        '''
        pool = Pool()
        StockMove = pool.get('stock.move')
        Uom = Pool().get('product.uom')

        if self.type != 'line':
            return
        if (not self.product and not self.product.kit and
                not self.product.kit_lines and
                self.product.explode_kit_in_sales):
            return
        if (shipment_type == 'out') != (self.quantity >= 0):
            return

        quantity = (self._get_move_quantity(shipment_type)
            - self._get_shipped_quantity(shipment_type))

        quantity = self.unit.round(quantity)
        if quantity <= 0:
            return

        if not self.sale.party.customer_location:
            raise PartyLocationError(
                gettext('sale.msg_sale_customer_location_required',
                    sale=self.sale.rec_name,
                    party=self.sale.party.rec_name))

        to_write, to_create = [], []
        depth = self.kit_depth + 1
        kit_lines = list(self.product.kit_lines)
        kit_lines = list(zip(kit_lines, [depth] * len(kit_lines)))
        print(-2, kit_lines)
        while kit_lines:
            kit_line = kit_lines.pop(0)
            depth = kit_line[1]
            kit_line = kit_line[0]
            move = kit_line.get_move(self, shipment_type)
            print(-1, move)
            move.on_change_product()
            move.save()
            print('move saved')
            if move:
                print(0, move._save_values)
                #to_create.append(move._save_values)
                to_create.append(move)
            product = kit_line.product
            print(1, move._save_values)
            print(2, product.default_uom_category.name)
            print(3, product.default_uom.name)
            print(5, move.uom.name)
            #print(6, move.product_uom_category.name)
            #print(7, move.uom.name)
            if product.kit_lines:
                product_kit_lines = list(product.kit_lines)
                product_kit_lines = list(zip(product_kit_lines,
                    [depth + 1] * len(product_kit_lines)))
                kit_lines = product_kit_lines + kit_lines

        print(10,self._save_values)
        if self._save_values:
            to_write.extend(([self], self._save_values))
            self.write(*to_write)
        print(20,StockMove.product.domain)
        #return StockMove.create(to_create)

    def get_moves_for_kit(self, shipment_type):
        '''
        Walk through the kit tree in depth-first order, get
        a sorted list with all the components of the kit and
        create moves for the kit components.

        s.a. sale/sale.py/get_move
        s.a. sale_kit/product.py/KitLine/get_move
        '''
        pool = Pool()
        StockMove = pool.get('stock.move')
        Uom = Pool().get('product.uom')

        if self.type != 'line':
            return
        if (not self.product and not self.product.kit and
                not self.product.kit_lines and
                self.product.explode_kit_in_sales):
            return
        if (shipment_type == 'out') != (self.quantity >= 0):
            return

        quantity = (self._get_move_quantity(shipment_type)
            - self._get_shipped_quantity(shipment_type))

        quantity = self.unit.round(quantity)
        if quantity <= 0:
            return

        if not self.sale.party.customer_location:
            raise PartyLocationError(
                gettext('sale.msg_sale_customer_location_required',
                    sale=self.sale.rec_name,
                    party=self.sale.party.rec_name))

        to_write, to_create = [], []
        depth = self.kit_depth + 1
        kit_lines = list(self.product.kit_lines)
        kit_lines = list(zip(kit_lines, [depth] * len(kit_lines)))
        print(-2, kit_lines)
        while kit_lines:
            kit_line = kit_lines.pop(0)
            depth = kit_line[1]
            kit_line = kit_line[0]
            #with Transaction().new_transaction() as new_transaction:
            #print('Sqlitecursor1', dir(Transaction().connection))
            #print('Sqlitecursor',
            #    dir(Transaction().connection.cursor().connection))
            #for cache in Transaction().connection.cursor().cache.values():
            #    print('cache1', cache)
            print('Trans', dir(Transaction()))
            Transaction().cache.clear()
            for item in to_create:
                print('itemc', item)
                item._local_cache.clear()
            move = kit_line.get_move(self, shipment_type)
            print(-1, move)
            #move.on_change_product()
            print('???',move.on_change_with_product_uom_category())
            if move:
                print(0, move._save_values)
                #to_create.append(move._save_values)
                to_create.append(move)
            #print('move', dir(move))
            print('move1', move._cache)
            move._cache.clear()
            print('move1', move._cache)
            print('move2', move._local_cache)
            for item in move._local_cache:
                print('itemlc', item)
            product = kit_line.product
            print(1, move._save_values)
            print(2, product.default_uom_category.name)
            print(3, product.default_uom.name)
            print(5, move.uom.name)
            move.save()
                #new_transaction.commit()
            print('move saved')
            print(6, move.product_uom_category.name)
            #print(7, move.uom.name)
            if product.kit_lines:
                product_kit_lines = list(product.kit_lines)
                product_kit_lines = list(zip(product_kit_lines,
                    [depth + 1] * len(product_kit_lines)))
                kit_lines = product_kit_lines + kit_lines

        print(10,self._save_values)
        if self._save_values:
            to_write.extend(([self], self._save_values))
            self.write(*to_write)
        print(20,StockMove.product.domain)
        return StockMove.create(to_create)

    @classmethod
    def create(cls, values):
        lines = super(SaleLine, cls).create(values)
        if (Transaction().context.get('explode_kit', True)
                and not Transaction().context.get('standalone', False)):
            lines.extend(cls.explode_kit(lines))
        return lines

    def get_kit_lines(self, kit_line=None):
        res = []
        if kit_line:
            childs = kit_line.kit_child_lines
        else:
            childs = self.kit_child_lines
        for kit_line in childs:
            res.append(kit_line)
            res += self.get_kit_lines(kit_line)
        return res

    @classmethod
    def write(cls, *args):
        actions = iter(args)
        to_write, to_reset, to_delete = [], [], []
        if (Transaction().context.get('explode_kit', True)
                and not Transaction().context.get('standalone', False)):
            for lines, values in zip(actions, actions):
                reset_kit = False
                # TODO Explode kit when add new line from standalone
                if ('product' in values or 'quantity' in values
                        or 'unit' in values):
                    reset_kit = True
                lines = lines[:]
                if reset_kit:
                    for line in lines:
                        to_delete += line.get_kit_lines()
                    lines = list(set(lines) - set(to_delete))
                    to_reset.extend(lines)
                to_write.extend((lines, values))
        else:
            to_write = args
        super(SaleLine, cls).write(*to_write)
        if to_delete:
            cls.delete(to_delete)
        if to_reset:
            cls.explode_kit(to_reset)

    @classmethod
    def copy(cls, lines, default=None):
        if default is None:
            default = {}
        default.setdefault('kit_child_lines', [])
        lines = [x for x in lines if not x.kit_parent_line]
        new_lines = super(SaleLine, cls).copy(lines, default=default)
        return new_lines

    def get_invoice_line(self):
        lines = super(SaleLine, self).get_invoice_line()
        for line in lines:
            line.sequence = self.sequence
        return lines

    def get_move(self, shipment_type):
        # Don't create a move for a kit when the kit is not exploded
        # Happens when shipment_type == 'return', be rather safe than sorry
        if self.type != 'line' or not self.product:
            return
        if self.product.kit and not self.product.explode_kit_in_sales:
            return
        return super(SaleLine, self).get_move(shipment_type)




# The COPYRIGHT file at the top level of this repository contains the full
# copyright notices and license terms.
from decimal import Decimal
from datetime import timedelta

from trytond.model import fields
from trytond.pool import Pool, PoolMeta
from trytond.pyson import Eval, Bool
from trytond.transaction import Transaction
from trytond.exceptions import UserError
from trytond.i18n import gettext

STATES = {
    'invisible': Bool(~Eval('kit')),
}
DEPENDS = ['kit']


class Product(metaclass=PoolMeta):
    __name__ = "product.product"
    explode_kit_in_sales = fields.Boolean('Expand components in sales',
        states=STATES, depends=DEPENDS,
        help='When activated, the components of the kit will be added as '
        'individual sale lines with their relative unit price and the kit '
        'itself will have set its unit price to 0.')
    components_in_description = fields.Boolean('Components in description',
        states=STATES, depends=DEPENDS,
        help='When activated, the components of the kit will be added to '
        'the sale line description.')

    @staticmethod
    def default_explode_kit_in_sales():
        return False

    @staticmethod
    def default_components_in_description():
        return True

    @classmethod
    def validate(cls, products):
        super(Product, cls).validate(products)
        for product in products:
            product.check_required_salable_products_in_kits()
            product.check_purchasable_in_kit()

    def check_required_salable_products_in_kits(self):
        KitLine = Pool().get('product.kit.line')

        if not self.kit:
            return

        not_salable_lines = KitLine.search_count([
                ('parent', 'in', [self.id]),
                ('product.salable', '=', False),
                ('parent.explode_kit_in_sales', '=', True),
                ])
        if not_salable_lines:
            raise UserError(gettext(
                    'sale_kit.salable_product_required_in_kit',
                    self.rec_name))

    def check_purchasable_in_kit(self):
        if self.kit and self.purchasable:
            raise UserError(gettext(
                    'sale_kit.kit_not_purchasable',
                    self.rec_name))


    @classmethod
    def get_sale_price(cls, products, quantity=0):
        pool = Pool()
        Uom = pool.get('product.uom')

        prices = {}

        uom = None
        if Transaction().context.get('uom'):
            uom = Uom(Transaction().context.get('uom'))

        non_kits = []
        for product in products:
            if not product.kit or product.kit_fixed_list_price:
                non_kits.append(product)
                continue
            if product.explode_kit_in_sales:
                prices[product.id] = Decimal('0.0')
                continue

            product_price = Decimal('0.0')
            for kit_line in product.kit_lines:
                with Transaction().set_context(uom=kit_line.unit):
                    product_price += (cls.get_sale_price([kit_line.product],
                            quantity=kit_line.quantity)[kit_line.product.id] *
                        Decimal(str(kit_line.quantity)))
            prices[product.id] = product_price

            if uom:
                prices[product.id] = Uom.compute_price(
                    product.default_uom, prices[product.id], uom)

        if non_kits:
            prices.update(super(Product, cls).get_sale_price(non_kits,
                    quantity))
        return prices

    def compute_shipping_date(self, date=None):
        Date = Pool().get('ir.date')

        if self.kit and self.components_in_description:
            shipping_date = Date.today()
            for kit_line in self.kit_lines:
                shipping_date = min(shipping_date,
                    kit_line.product.compute_shipping_date(date))
        else:
            shipping_date = super(Product, self).compute_shipping_date(
                date=date)
        return shipping_date


class ProductKitLine(metaclass=PoolMeta):
    __name__ = 'product.kit.line'

    def get_sale_price(self):
        parent = self.parent
        if parent.kit_fixed_list_price:
            return False
        parent_kit_lines = self.search([
                ("product", "=", parent.id),
                ])
        for line in parent_kit_lines:
            if line in [x for x in line.product.kit_lines]:
                return line.get_sale_price()
        return True

    @classmethod
    def validate(cls, lines):
        super(ProductKitLine, cls).validate(lines)
        for line in lines:
            line.check_required_salable_lines()

    def check_required_salable_lines(self):
        if (self.parent.explode_kit_in_sales and
                not self.product.salable):
            raise UserError(gettext('sale_kit.salable_lines_required'))

    def get_move(self, sale_line, shipment_type):
        '''
        Return moves for the kit line according to shipment_type

        Adapted from SaleLine/get_move
        '''
        pool = Pool()
        Move = pool.get('stock.move')
        Product = pool.get('product.product')
        Uom = pool.get('product.uom')

        if sale_line.type != 'line':
            return
        if not sale_line.product:
            return
        if sale_line.product.type == 'service':
            return

        if (shipment_type == 'out') != (sale_line.quantity >= 0):
            return

        to_ship = sale_line.quantity * self.quantity
        shipped = sale_line._get_shipped_quantity_for_kit_line(self)
        quantity = to_ship - shipped

        if quantity <= 0:
            return

        if not sale_line.sale.party.customer_location:
            raise UserError(gettext(
                        'sale_kit.customer_location_required',
                        sale=sale_line.sale.rec_name,
                        line=sale_line.rec_name,
                        ))

        product_price = Decimal('0.0')
        with Transaction().set_context(uom=self.unit):
            product_price += (Product.get_sale_price([self.product],
                    quantity=self.quantity)[self.product.id] *
                    Decimal(str(self.quantity)))

        move = Move()
        move.quantity = quantity
        move.uom = self.product.default_uom
        print('YYYYYYYYYYY', move.uom.name)
        move.product = self.product
        move.from_location = sale_line.from_location
        move.to_location = sale_line.to_location
        move.state = 'draft'
        move.company = sale_line.sale.company
        move.unit_price = product_price
        move.currency = sale_line.sale.currency
        move.planned_date = sale_line.shipping_date
        move.invoice_lines = sale_line.invoice_lines
        move.origin = sale_line
        return move


class ProductSupplier(metaclass=PoolMeta):
    __name__ = 'purchase.product_supplier'

    def compute_supply_date(self, date=None):
        Date = Pool().get('ir.date')

        product = self.product
        if product.kit and product.components_in_description:
            lead_time = timedelta(0)
            for kit_line in product.kit_lines:
                lead_time = max(lead_time, kit_line.product.lead_time)
            if not date:
                date = Date.today()
            lead_time = date + lead_time
        else:
            lead_time = super(ProductSupplier, self).compute_supply_date(
                date=date)
        return lead_time

    def compute_purchase_date(self, date):
        Date = Pool().get('ir.date')

        product = self.product
        if product.kit and product.components_in_description:
            lead_time = timedelta(0)
            for kit_line in product.kit_lines:
                lead_time = max(lead_time, kit_line.product.lead_time)
            if not date:
                date = Date.today()
            lead_time = date - lead_time
        else:
            lead_time = super(ProductSupplier, self).compute_purchase_date(
                date=date)
        return lead_time



# This file is part of sale_kit module for Tryton.
# The COPYRIGHT file at the top level of this repository contains
# the full copyright notices and license terms.
from decimal import Decimal

from trytond.model import fields
from trytond.pool import Pool, PoolMeta
from trytond.pyson import Eval, Bool
from trytond.transaction import Transaction
from trytond.i18n import gettext
from trytond.exceptions import UserError

__all__ = ['Product', 'ProductKitLine']
STATES = {
    'invisible': Bool(~Eval('kit')),
}
DEPENDS = ['kit']


class Product(metaclass=PoolMeta):
    __name__ = "product.product"
    explode_kit_in_sales = fields.Boolean('Explode in Sales', states=STATES,
            depends=DEPENDS)

    @staticmethod
    def default_explode_kit_in_sales():
        return True

    @staticmethod
    def default_kit_fixed_list_price():
        return True

    @classmethod
    def validate(cls, products):
        super(Product, cls).validate(products)
        for product in products:
            product.check_required_salable_products_in_kits()

    def check_required_salable_products_in_kits(self):
        KitLine = Pool().get('product.kit.line')

        if not self.kit:
            return

        n_not_salable_lines = KitLine.search_count([
                ('parent', 'in', [self.id]),
                ('product.salable', '=', False),
                ('parent.explode_kit_in_sales', '=', True),
                ])
        if n_not_salable_lines:
            raise UserError(gettext('sale_kit.salable_product_required_in_kit',
                product=self.rec_name))

    @classmethod
    def get_sale_price(cls, products, quantity=0):
        '''
        Return the sale price for products and quantity.
        It uses if exists from the context:
            uom: the unit of measure
            currency: the currency id for the returned price
        '''
        pool = Pool()
        Uom = pool.get('product.uom')
        User = pool.get('res.user')
        Currency = pool.get('currency.currency')
        Date = pool.get('ir.date')

        today = Date.today()
        prices = {}

        uom = None
        if Transaction().context.get('uom'):
            uom = Uom(Transaction().context.get('uom'))

        currency = None
        if Transaction().context.get('currency'):
            currency = Currency(Transaction().context.get('currency'))

        user = User(Transaction().user)

        todo_products = []
        for product in products:
            if not product.kit or product.kit_fixed_list_price:
                todo_products.append(product)
                continue
            if product.explode_kit_in_sales:
                prices[product.id] = Decimal('0.0')
                continue

            product_price = Decimal('0.0')
            for kit_line in product.kit_lines:
                with Transaction().set_context(uom=kit_line.unit):
                    product_price += (cls.get_sale_price([kit_line.product],
                            quantity=kit_line.quantity)[kit_line.product.id] *
                        Decimal(str(kit_line.quantity)))
            prices[product.id] = product_price

            if uom:
                prices[product.id] = Uom.compute_price(
                    product.default_uom, prices[product.id], uom)
            if currency and user.company:
                if user.company.currency != currency:
                    date = Transaction().context.get('sale_date') or today
                    with Transaction().set_context(date=date):
                        prices[product.id] = Currency.compute(
                            user.company.currency, prices[product.id],
                            currency, round=False)

        if todo_products:
            prices.update(super(Product, cls).get_sale_price(todo_products,
                    quantity))

        return prices


class ProductKitLine(metaclass=PoolMeta):
    __name__ = 'product.kit.line'


    def get_sale_price(self):
        parent = self.parent
        if parent.kit_fixed_list_price:
            return False
        parent_kit_lines = self.search([
                ("product", "=", parent.id),
                ])
        for line in parent_kit_lines:
            if line in [x for x in line.product.kit_lines]:
                return line.get_sale_price()
        return True

    @classmethod
    def validate(cls, lines):
        super(ProductKitLine, cls).validate(lines)
        for line in lines:
            line.check_required_salable_lines()

    def check_required_salable_lines(self):
        if (self.parent.explode_kit_in_sales and
                not self.product.salable):
            raise UserError(gettext('sale_kit.salable_lines_required'))
